/*
  Copyright (c) 2015, Matthias Schiffer <mschiffer@universe-factory.net>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#pragma once

#include "item.hpp"

#include <stdexcept>
#include <set>
#include <map>


namespace solar {

class generator_t {
public:
	class conflict_error : public std::runtime_error {
	public:
		explicit conflict_error(const std::string& what_arg) : std::runtime_error(what_arg) {}
		explicit conflict_error(const char* what_arg) : std::runtime_error(what_arg) {}
	};

private:
	std::vector<std::tuple<item_t, std::vector<std::string>, std::string>> rules;
	std::map<item_t, size_t> rule_ids;
	std::multimap<std::string, size_t> nonterms;

	std::set<std::string> nonterminals;
	std::set<symbol_t> terminals;

	std::multimap<symbol_t, item_t> items;
	std::map<std::set<item_t>, size_t> itemsets;

	std::map<std::pair<size_t, symbol_t>, size_t> shifts;
	std::map<size_t, size_t> reductions;
	std::map<std::pair<size_t, std::string>, size_t> gotos;

	std::set<size_t> shift_conflicts;

	std::map<std::string, std::string> nonterm_types;
	std::map<symbol_t, std::pair<std::string, std::string>> term_types;

	void close_set(std::set<item_t> *set);
	std::set<item_t> get_set(const std::string &nonterm);

	std::pair<std::map<std::set<item_t>, size_t>::iterator, bool> add_set(const std::set<item_t> &set) {
		return itemsets.emplace(set, itemsets.size());
	}

	void add_reduction(size_t from, size_t to) {
		if (reductions.count(from))
			throw conflict_error("reduce/reduce conflict");
		if (shift_conflicts.count(from))
			throw conflict_error("shift/reduce conflict");

		reductions.emplace(from, to);
	}

	void add_shift(size_t from, const symbol_t &sym, size_t to) {
		if (reductions.count(from))
			throw conflict_error("shift/reduce conflict");

		shift_conflicts.insert(from);
		shifts.emplace(std::make_pair(from, sym), to);
	}

	void add_goto(size_t from, const std::string &nonterm, size_t to) {
		gotos.emplace(std::make_pair(from, nonterm), to);
	}

	void generate_itemsets();

public:
	const std::string & get_nonterm_type(const std::string &sym) const {
		static const std::string empty;

		auto it = nonterm_types.find(sym);
		if (it == nonterm_types.end())
			return empty;
		else
			return it->second;
	}

	const std::pair<std::string, std::string> & get_term_type(const symbol_t &sym) const {
		static const std::pair<std::string, std::string> empty;

		auto it = term_types.find(sym);
		if (it == term_types.end())
			return empty;
		else
			return it->second;
	}

	const std::string & get_type(const symbol_t &sym) const {
		switch (sym.get_type()) {
		case SYMBOL_TYPE_NONTERM:
			return get_nonterm_type(sym.get_value());

		default:
			return get_term_type(sym).first;
		}
	}

	const std::set<std::string> & get_nonterminals() const {
		return nonterminals;
	}

	const std::set<symbol_t> & get_terminals() const {
		return terminals;
	}

	size_t get_state_count() const {
		return itemsets.size();
	}

	const std::vector<std::tuple<item_t, std::vector<std::string>, std::string>> & get_rules() const {
		return rules;
	}

	const std::map<size_t, size_t> & get_reductions() const {
		return reductions;
	}

	const std::map<std::pair<size_t, symbol_t>, size_t> & get_shifts() const {
		return shifts;
	}

	const std::map<std::pair<size_t, std::string>, size_t> & get_gotos() const {
		return gotos;
	}

	generator_t(const std::vector<std::tuple<item_t, std::vector<std::string>, std::string>> &rules0,
		    const std::map<std::string, std::string> &nonterm_types0,
		    const std::map<symbol_t, std::pair<std::string, std::string>> &term_types0);
};

}
